package jp.ac.hokudai.ist.main.aiwww.stlics.solver.fuzzy;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import jp.ac.hokudai.ist.main.aiwww.stlics.problem.Constraint;
import jp.ac.hokudai.ist.main.aiwww.stlics.problem.Domain;
import jp.ac.hokudai.ist.main.aiwww.stlics.problem.Problem;
import jp.ac.hokudai.ist.main.aiwww.stlics.problem.Variable;
import jp.ac.hokudai.ist.main.aiwww.stlics.solver.AbstractFuzzySolver;
import jp.ac.hokudai.ist.main.aiwww.stlics.solver.DomainPruner;
import jp.ac.hokudai.ist.main.aiwww.stlics.solver.trait.SystematicSolver;
import jp.ac.hokudai.ist.main.aiwww.stlics.util.AssignmentList;

/**
 * ファジィCSP用の前方チェック法を実装したクラスです．
 * オプションの指定により，最小ドメイン優先ヒューリスティック(MRV)を使用することもできます．
 * 枝刈としてドメインの要素をhideするので，変数ごとに独自のドメインを持つ必要があります．
 * 多項制約を持つ問題に対しても，前方チェックを行います．
 * @author Takuto Yanagida
 * @version 2010/11/17
 */
public class FuzzyForwardChecking extends AbstractFuzzySolver implements SystematicSolver {

	private static enum BranchCode {CONTINUE, TERMINATE}

	private final AssignmentList solution_ = new AssignmentList();
	private final Variable[] vars_;
	private Constraint[][][] relCons_;  // 2変数間の制約をキャッシュするテーブル
	private double solutionWorstDeg_ = 0.0;  // デフォルトは0.0

	private int iterCount_;
	private long endTime_;
	private boolean useMRV_ = false;
	private double degInc_ = 0.0;

	private final Variable[] sequence_;
	private Constraint[] unaryCons_;
	private final boolean[] checkedCons_;
	private boolean pruneIntensively_ = false;

	/**
	 * ファジィ制約充足問題を指定してソルバを生成します．
	 * @param p ファジィ制約充足問題
	 */
	public FuzzyForwardChecking(Problem p) {
		super(p, "Forward checking for fuzzy CSPs");
		vars_ = pro_.variables().toArray(new Variable[pro_.variableSize()]);
		sequence_ = new Variable[pro_.variableSize()];
		initializeRelatedConstraintTable();
		checkedCons_ = new boolean[pro_.constraintSize()];

		List<Constraint> temp = new ArrayList<>();
		for(Constraint c: pro_.constraints()) {
			if(c.size() == 1) temp.add(c);
		}
		unaryCons_ = temp.toArray(new Constraint[temp.size()]);  // 空でも作ること
	}

	/**
	 * ファジィ制約充足問題と最悪充足度を指定してソルバを生成します．
	 * 最悪充足度に従って枝刈りを開始します．
	 * もし，最悪の充足度が予想できる場合は枝刈りの効率が上がり，解を速く得られるようになりますが，
	 * この値以下の解しか存在しないとき，アルゴリズムは解を見つけることなく停止します．
	 * @param p ファジィ制約充足問題
	 * @param worstSatisfactionDegree 最悪充足度
	 */
	public FuzzyForwardChecking(Problem p, double worstSatisfactionDegree) {
		this(p);
		solutionWorstDeg_ = worstSatisfactionDegree;
	}

	// 2変数間の制約をキャッシュするテーブルを初期化する．
	private void initializeRelatedConstraintTable() {
		List<Constraint> temp = new ArrayList<>();
		relCons_ = new Constraint[vars_.length][vars_.length][];
		for(int j = 0; j < vars_.length; ++j) {
			relCons_[j] = new Constraint[vars_.length][];
			for(int i = 0; i < vars_.length; i++) {
				if(i < j) {
					pro_.constraintsBetween(vars_[i], vars_[j], temp);
					relCons_[j][i] = temp.toArray(new Constraint[0]);
				}
			}
		}
	}

	// 2変数間の制約をキャッシュするテーブルをから制約の配列を取得する．
	private Constraint[] getConstraintsBetween(int vi_index, int vj_index) {
		if(vi_index < vj_index) return relCons_[vj_index][vi_index];
		return relCons_[vi_index][vj_index];
	}

	// 単項制約を現在の最悪充足度より悪化させるドメインの要素を枝刈りする．
	private boolean pruneUnaryConstraints() {
		for(Constraint c: unaryCons_) {
			final Variable v = c.at(0);
			final int orgVal = v.value();  // 値を保存
			final Domain d = v.domain();
			final DomainPruner dc = (DomainPruner)v.solverObject;

			for(int i = 0, n = d.size(); i < n; ++i) {
				v.assign(d.at(i));
				if(c.satisfactionDegree() <= solutionWorstDeg_) {
					dc.hide(i, -1);  // ここで枝刈り!!
				}
			}
			v.assign(orgVal);  // 値を復元
			if(dc.isEmpty()) return false;
		}
		return true;
	}

	// 現在変数と一つの未来変数との整合を調べ，整合しないドメインの要素を枝刈りする(制約のスコープの未割り当て変数が一つの場合)．
	private boolean checkForwardConsistency(int currentLevel, Variable vi, Constraint c) {
		Domain di = vi.domain();
		DomainPruner dci = (DomainPruner)vi.solverObject;
		for(int i = 0, n = di.size(); i < n; ++i) {
			if(dci.isValueHidden(i)) continue;
			vi.assign(di.at(i));
			if(c.satisfactionDegree() <= solutionWorstDeg_) {  // 「小なりイコール」のときは解とならない※
				dci.hide(i, currentLevel);  // ここで枝刈り!!
			}
		}
		vi.clear();
		return !dci.isEmpty();  // 未来変数viのドメインdiが空にならなければ成功．
	}

	// 現在変数と一つの未来変数との整合を調べ，整合しないドメインの要素を枝刈りする(制約のスコープの未割り当て変数が二つの場合)．
	private boolean checkForwardConsistency2(int currentLevel, Variable vi, Constraint c) {
		Domain di = vi.domain();
		DomainPruner dci = (DomainPruner)vi.solverObject;
		Variable vj = null;
		for(int i = 0, n = c.size(); i < n; ++i) {
			Variable v = c.at(i);
			if(v.isEmpty() && v != vi) {
				vj = v;
				break;
			}
		}
		Domain dj = vj.domain();
		DomainPruner dcj = (DomainPruner)vj.solverObject;
		loop_i: for(int i = 0, ni = di.size(); i < ni; ++i) {
			if(dci.isValueHidden(i)) continue;
			vi.assign(di.at(i));  // viへの仮割り当て
			for(int j = 0, nj = dj.size(); j < nj; ++j) {
				if(dcj.isValueHidden(j)) continue;
				vj.assign(dj.at(j));  // viへの仮割り当て
				double s = c.satisfactionDegree();
				if(s > solutionWorstDeg_) continue loop_i;  // viへの仮割り当てはOKだった→次の仮割り当てへ
			}
			dci.hide(i, currentLevel);  // 「小なりイコール」のときは解とならない※
		}
		vj.clear();
		vi.clear();
		return !dci.isEmpty();  // 未来変数viのドメインdiが空にならなければ成功．
	}

	// 現在変数と一つの未来変数との整合を調べ，整合しないドメインの要素を枝刈りする(制約のスコープの未割り当て変数が三つの場合)．
	private boolean checkForwardConsistency3(int currentLevel, Variable vi, Constraint c) {
		Domain di = vi.domain();
		DomainPruner dci = (DomainPruner)vi.solverObject;
		Variable vj = null, vk = null;
		for(int i = 0, n = c.size(); i < n; ++i) {
			Variable v = c.at(i);
			if(v.isEmpty() && v != vi) {
				if(vj == null) {
					vj = v;
				} else {
					vk = v;
					break;
				}
			}
		}
		Domain dj = vj.domain(), dk = vk.domain();
		DomainPruner dcj = (DomainPruner)vj.solverObject, dck = (DomainPruner)vk.solverObject;
		loop_i: for(int i = 0, ni = di.size(); i < ni; ++i) {
			if(dci.isValueHidden(i)) continue;
			vi.assign(di.at(i));  // viへの仮割り当て
			for(int j = 0, nj = dj.size(); j < nj; ++j) {
				if(dcj.isValueHidden(j)) continue;
				vj.assign(dj.at(j));  // viへの仮割り当て
				for(int k = 0, nk = dk.size(); k < nk; ++k) {
					if(dck.isValueHidden(k)) continue;
					vk.assign(dk.at(k));  // viへの仮割り当て
					double s = c.satisfactionDegree();
					if(s > solutionWorstDeg_) continue loop_i;  // viへの仮割り当てはOKだった→次の仮割り当てへ
				}
			}
			dci.hide(i, currentLevel);  // 「小なりイコール」のときは解とならない※
		}
		vk.clear();
		vj.clear();
		vi.clear();
		return !dci.isEmpty();  // 未来変数viのドメインdiが空にならなければ成功．
	}

	// 多項制約の場合で，かつ，未割り当て変数が四つ以上あるときに，未割り当て変数の割り当ての全ての組合せを調べて，枝刈りする．
	private boolean checkForwardConsistencyN(int currentLevel, Variable vi, Constraint c, int emptySize) {
		Domain di = vi.domain();
		DomainPruner dci = (DomainPruner)vi.solverObject;
		Variable[] emp = new Variable[emptySize - 1];
		int j = 0;
		for(int i = 0, n = c.size(); i < n; ++i) {
			Variable v = c.at(i);
			if(v.isEmpty() && v != vi) emp[j++] = v;
		}
		int[] indexes = new int[emp.length];

		loop_i: for(int i = 0, n = di.size(); i < n; ++i) {
			if(dci.isValueHidden(i)) continue;
			vi.assign(di.at(i));  // viへの仮割り当て
			Arrays.fill(indexes, 0);

			comLoop: while(true) {
				boolean hidden = false;
				for(int k = 0; k < emp.length; ++k) {
					Domain dk = emp[k].domain();
					DomainPruner dck = (DomainPruner)emp[k].solverObject;
					if(dck.isValueHidden(indexes[k])) {
						hidden = true;
						break;
					}
					emp[k].assign(dk.at(indexes[k]));
				}
				if(!hidden) {
					double s = c.satisfactionDegree();
					if(s > solutionWorstDeg_) continue loop_i;  // viへの仮割り当てはOKだった→次の仮割り当てへ
				}
				for(int k = 0; k < emp.length; ++k) {
					indexes[k]++;
					if(indexes[k] < emp[k].domain().size()) break;
					indexes[k] = 0;
					if(k == emp.length - 1) break comLoop;
				}
			}
			dci.hide(i, currentLevel);
		}
		for(Variable v: emp) v.clear();
		vi.clear();
		return !dci.isEmpty();  // 未来変数viのドメインdiが空にならなければ成功．
	}

	// 現在変数の割り当てから未来変数に割り当て可能かを調べる．
	private boolean checkForward(final int currentLevel, final int currentIndex) {
		for(Variable v_i: vars_) {
			if(!v_i.isEmpty()) continue;  // 過去変数か現在変数だったら
			Constraint[] cs = getConstraintsBetween(currentIndex, v_i.index());
			for(Constraint c: cs) {
				int emptySize = c.emptyVariableSize();
				if(emptySize == 1) {
					if(!checkForwardConsistency(currentLevel, v_i, c)) return false;
				} else if(pruneIntensively_) {  // オプションによる
					if(emptySize == 2) {
						if(!checkForwardConsistency2(currentLevel, v_i, c)) return false;
					} else if(emptySize == 3) {
						if(!checkForwardConsistency3(currentLevel, v_i, c)) return false;
					} else if(emptySize > 3) {
						if(!checkForwardConsistencyN(currentLevel, v_i, c, emptySize)) return false;
					}
				}
			}
		}
		return true;
	}

	// 現在変数の割り当てから過去変数との充足度が現在の最悪充足度より悪化していないかを調べる．
	private boolean checkBackwardConsistency(Variable vc) {
		Arrays.fill(checkedCons_, false);  // 再利用
		for(int i = 0; i < vars_.length; ++i) {  // 過去変数を探す
			Variable vi = vars_[i];
			if(vi == vc || vi.isEmpty()) continue;  // 未来変数か現在変数だったら
			Constraint[] cs = getConstraintsBetween(vc.index(), i);
			for(Constraint c: cs) {
				if(checkedCons_[c.index()]) continue;  // 多項制約では重複の恐れがあるため
				double s = c.satisfactionDegree();
				if(s != Constraint.UNDEFINED && s <= solutionWorstDeg_) {  // 「小なりイコール」のときは解とならない
					return false;
				}
				checkedCons_[c.index()] = true;
			}
		}
		return true;
	}

	private void refresh() {
		for(int i = 0; i < sequence_.length; ++i) {
			int index_vi = sequence_[i].index();
			for(int j = i + 1; j < sequence_.length; ++j) {
				Variable vj = sequence_[j];
				Constraint[] cs = getConstraintsBetween(index_vi, vj.index());
				for(Constraint c: cs) {
					int orgVal = vj.value();
					Domain dj = vj.domain();
					DomainPruner dcj = (DomainPruner)vj.solverObject;
					for(int k = 0, n = dj.size(); k < n; ++k) {
						if(dcj.isValueHidden(k)) continue;
						vj.assign(dj.at(k));
						if(c.satisfactionDegree() <= solutionWorstDeg_) {
							dcj.hide(k, i);  // ここで枝刈り!!
						}
					}
					vj.assign(orgVal);
				}
			}
		}
	}

	// もっともドメインの小さな変数のインデックスを返す．
	private int indexOfVariableWithMRV() {
		int index = 0;
		int size = Integer.MAX_VALUE;
		for(int i = 0; i < vars_.length; ++i) {
			final Variable v = vars_[i];
			if(!v.isEmpty()) continue;
			final Domain d = v.domain();
			final int s = d.size() - ((DomainPruner)v.solverObject).hiddenSize();
			if(s < size) {
				size = s;
				index = i;
			}
		}
		return index;
	}

//
// 下のコメントアウトはしばらく消さないこと!!
//

//	// 探索の終了を調べる
//	private BranchCode checkTermination(int currentLevel) {
//		if(iterLimit_ < iterCount_++ || endTime_ < System.currentTimeMillis()) {
//			if(targetDeg_ != UNSPECIFIED && targetDeg_ <= solutionWorstDeg_) return BranchCode.SUCCESS;  // 違反度が目標より改善された→探索終了
//			if(targetDeg_ == UNSPECIFIED && 0.0 < solutionWorstDeg_) return BranchCode.SUCCESS;
//			return BranchCode.TERMINATE;  // 規定回数繰り返したか制限時間を超えた→探索終了
//		}
//		if(currentLevel != vars_.length - 1) {
//			return BranchCode.CONTINUE;  // まだ最後まで行っていない→探索継続
//		}
//		double s = pro_.worstSatisfactionDegree();
//		// ↓があるのは，一度リフレッシュした後に枝刈りを更新せずに探索を継続するから！！
//		if(s > solutionWorstDeg_) {  // 「大なり」のときは新たな解とする※
//			solutionWorstDeg_ = s;
//			solution_.set(pro_);
//			if(debug) debugStream.println("\tRefreshed: " + solutionWorstDeg_);
//			if(targetDeg_ != UNSPECIFIED && targetDeg_ <= solutionWorstDeg_) return BranchCode.SUCCESS;  // 違反度が目標より改善された→探索終了
//			return BranchCode.FINISHED_FAILUER_UPDATED;  // 解は発見したものの既出解以下だったがwsdを更新した→探索継続
//		}
//		return BranchCode.FINISHED_FAILUER;  // 解は発見したものの既出解以下だった→探索継続
//	}

//	// 探索を1変数ずつ行う．
//	private BranchCode branch(int currentLevel) {
//		BranchCode bc = BranchCode.CONTINUE;
//		final int vc_index = useMRV_ ? indexOfVariableWithMRV() : currentLevel;
//		final Variable vc = vars_[vc_index];
//		final Domain d = vc.domain();
//		final DomainPruner dc = (DomainPruner)vc.solverObject;
//		sequence_[currentLevel] = vc;
//
//		for(int i = 0, n = d.size(); i < n; ++i) {
//			if(dc.isValueHidden(i)) continue;
//			vc.assign(d.at(i));
//
//			for(Variable v: vars_) ((DomainPruner)v.solverObject).reveal(currentLevel);
//			if(!checkBackwardConsistency(vc)) continue;
//			if(!checkForward(currentLevel, vc_index)) continue;
//
//			bc = checkTermination(currentLevel);
//			if(bc == BranchCode.SUCCESS || bc == BranchCode.TERMINATE) break;
//
//			switch(bc) {
//			case FINISHED_FAILUER_UPDATED:
//				pruneUnaryConstraints();
//				refresh();
//			case FINISHED_FAILUER: continue;
//			}
//
//			bc = branch(currentLevel + 1);
//			if(bc == BranchCode.SUCCESS || bc == BranchCode.TERMINATE) break;
//		}
//		vc.clear();
//		return bc;
//	}

	// 探索を1変数ずつ行う．
	private BranchCode branch(int currentLevel) {
		BranchCode bc = BranchCode.CONTINUE;
		final int vc_index = useMRV_ ? indexOfVariableWithMRV() : currentLevel;
		final Variable vc = vars_[vc_index];
		final Domain d = vc.domain();
		final DomainPruner dc = (DomainPruner)vc.solverObject;
		sequence_[currentLevel] = vc;

		for(int i = 0, n = d.size(); i < n; ++i) {
			if(dc.isValueHidden(i)) continue;
			if(iterLimit_ < iterCount_++ || endTime_ < System.currentTimeMillis()) {
				bc = BranchCode.TERMINATE;  // 制限により探索終了
				break;
			}
			vc.assign(d.at(i));

			for(Variable v: vars_) ((DomainPruner)v.solverObject).reveal(currentLevel);
			if(!checkBackwardConsistency(vc)) continue;
			if(!checkForward(currentLevel, vc_index)) continue;

			int nextLevel = currentLevel + 1;
			bc = (nextLevel == vars_.length - 1) ? branchLast(nextLevel) : branch(nextLevel);
			if(bc == BranchCode.TERMINATE) break;
		}
		if(bc == BranchCode.CONTINUE) {  // 親に戻って探索するときは，ここでの枝刈りを元に戻す．
			for(Variable v: vars_) ((DomainPruner)v.solverObject).reveal(currentLevel);
		}
		vc.clear();
		return bc;
	}

	// 探索を最後の変数で行う．
	private BranchCode branchLast(int currentLevel) {
		BranchCode bc = BranchCode.CONTINUE;
		final Variable vc = vars_[useMRV_ ? indexOfVariableWithMRV() : currentLevel];
		final Domain d = vc.domain();
		final DomainPruner dc = (DomainPruner)vc.solverObject;
		sequence_[currentLevel] = vc;

		for(int i = 0, n = d.size(); i < n; ++i) {
			if(dc.isValueHidden(i)) continue;
			if(iterLimit_ < iterCount_++ || endTime_ < System.currentTimeMillis()) {
				bc = BranchCode.TERMINATE;  // 制限により探索終了
				break;
			}
			vc.assign(d.at(i));

			double deg = pro_.worstSatisfactionDegree();
			if(deg > solutionWorstDeg_) {  // 「大なり」のときは新たな解とする
				solutionWorstDeg_ = deg;
				solution_.set(pro_);
				bc = BranchCode.TERMINATE;
				if(targetDeg_ != UNSPECIFIED && targetDeg_ <= solutionWorstDeg_) break;  // 目標到達により探索終了
				pruneUnaryConstraints();
				refresh();
			}
		}
		vc.clear();
		return bc;
	}

	// 探索を行う．
	@Override
	protected boolean exec() {
		endTime_ = (timeLimit_ == -1) ? Long.MAX_VALUE : (System.currentTimeMillis() + timeLimit_);
		iterCount_ = 0;

		for(Variable v: vars_) v.solverObject = new DomainPruner(v.domain().size());  // 枝刈り器の生成
		pro_.clearAllVariables();
		if(!pruneUnaryConstraints()) return false;  // worstSatisfactionDegree_が更新されたので呼び出す

		boolean success = false;
		while(!(iterLimit_ < iterCount_++ || endTime_ < System.currentTimeMillis())) {
			branch(0);
			if(solution_.isEmpty()) break;
			if(debug) debugStream.println("\tFound a solution: " + solutionWorstDeg_);
			if(targetDeg_ == UNSPECIFIED) {  // 充足度指定なし
				success = true;
				solutionWorstDeg_ += degInc_;  // 制限内で次の解を探す
			} else if(targetDeg_ <= solutionWorstDeg_) {  // 充足度指定あり，かつそれを上回った
				success = true;
				break;
			}
			for(Variable v: vars_) ((DomainPruner)v.solverObject).revealAll();
		}
		if(success) solution_.apply();
		for(Variable v: vars_) v.solverObject = null;  // 枝刈り器の削除
		return success;
	}

	/**
	 * ソルバの停止の条件となる達成目標として，制約充足度を設定します．
	 * 指定した充足度以上になると，ソルバは成功として停止します．
	 * デフォルト(未設定)は0.8です．
	 * @param rate 割合．UNSPECIFIEDは未設定を表す．
	 */
	@Override
	public void setTargetRate(double rate) {
		targetDeg_ = rate;
		if(targetDeg_ == UNSPECIFIED) {
			solutionWorstDeg_ = 0.0;
		} else {
			// targetDegree_よりもわずかに小さいworstSatisfactionDegree_を求める
			double e = Double.MIN_NORMAL;
			solutionWorstDeg_ = targetDeg_ - e;
			while(solutionWorstDeg_ >= targetDeg_) {
				e *= 10;
				solutionWorstDeg_ = targetDeg_ - e;
			}
		}
	}

	/**
	 * 最小ドメイン優先ヒューリスティック(MRV)を使用するかどうかを指定します．
	 * 問題によってはMRVの使用が処理時間を長くすることがあります．
	 * デフォルトではfalseです．
	 * @param flag MRVを使うかどうか．trueなら使う．falseなら使わない．
	 */
	public void setUsingMinimumRemainingValuesHeuristics(boolean flag) {
		useMRV_ = flag;
	}

	/**
	 * 解が見つかり，探索を継続する場合，最悪制約充足度をどれだけ増加させるかを指定します．
	 * @param degree 増加させる制約充足度
	 */
	public void setIncrementStepOfWorstSatisfactionDegree(double degree) {
		degInc_ = degree;
	}

	/**
	 * 問題に3項以上の制約が含まれていたときに，積極的に枝刈りを行うかどうかを指定します．
	 * 問題によっては積極的な枝刈りが処理時間を長くすることがあります．
	 * デフォルトではfalseです．
	 * @param flag 積極的に枝刈りを行うかどうか．trueなら行う．falseなら行わない．
	 */
	public void setIntensivePruning(boolean flag) {
		pruneIntensively_ = flag;
	}

}
